//----------------------------------------------------------------------
//  Author: Thomas S. Ullrich (Feb 2005)
//
//  This file is part of hxBrowser.
//  
//  Description: Definition and implementation of the StThreeVector 
//               class (taken from StarClassLibrary)
//----------------------------------------------------------------------
#ifndef ST_THREE_VECTOR_HH
#define ST_THREE_VECTOR_HH

#include <cmath>
#include <iostream>

using namespace std;

template<class T = double> class StThreeVector {
public:    
    StThreeVector(T = 0, T = 0, T = 0);
    ~StThreeVector();

    StThreeVector(const StThreeVector<float>&);
    StThreeVector(const StThreeVector<double>&);
    
    StThreeVector(const float*); 
    StThreeVector(const double*);
    
    StThreeVector<T>& operator=(const StThreeVector<float>&);
    StThreeVector<T>& operator=(const StThreeVector<double>&);
    
    void setX(T);
    void setY(T);
    void setZ(T);

    void setPhi(T);
    void setTheta(T);
    void setMag(T);
    void setMagnitude(T);
    
    T   x()                        const;
    T   y()                        const;
    T   z()                        const;
    T   theta()                    const;
    T   cosTheta()                 const;
    T   phi()                      const;
    T   perp()                     const;
    T   perp2()                    const;
    T   magnitude()                const;
    T   mag()                      const;
    T   mag2()                     const;
    T   pseudoRapidity()           const;
    T   operator() (size_t)        const;
    T   operator[] (size_t)        const;

    T&  operator() (size_t);
    T&  operator[] (size_t);
    
    T   massHypothesis(T mass)     const;
    
    StThreeVector<T>  unit()       const;
    StThreeVector<T>  orthogonal() const;

    void  rotateX(T);
    void  rotateY(T);
    void  rotateZ(T);
    
    StThreeVector<T>  operator- ();
    StThreeVector<T>  operator+ ();
    StThreeVector<T>& operator*= (double);
    StThreeVector<T>& operator/= (double);
    StThreeVector<T>  pseudoProduct(double,double,double) const;
 
    T                angle(const StThreeVector<float>&) const;
    StThreeVector<T> cross(const StThreeVector<float>&) const;
    T                dot  (const StThreeVector<float>&) const;
    StThreeVector<T> pseudoProduct(const StThreeVector<float>&) const;
    
    T                angle(const StThreeVector<double>&) const;
    T                dot  (const StThreeVector<double>&) const;
    StThreeVector<T> cross(const StThreeVector<double>&) const;
    StThreeVector<T> pseudoProduct(const StThreeVector<double>&) const;

    bool operator == (const StThreeVector<float>& v) const;
    bool operator != (const StThreeVector<float>& v) const;
    StThreeVector<T>& operator+= (const StThreeVector<float>&);
    StThreeVector<T>& operator-= (const StThreeVector<float>&);
    
    bool operator == (const StThreeVector<double>& v) const;
    bool operator != (const StThreeVector<double>& v) const;
    StThreeVector<T>& operator+= (const StThreeVector<double>&);
    StThreeVector<T>& operator-= (const StThreeVector<double>&);

    int             valid(double world = 1.e+5) const;

protected:
    T    mX1, mX2, mX3;
};

//
//        Implementation of member functions
//
template<class T>
inline StThreeVector<T>::StThreeVector(T x, T y, T z)
    : mX1(x), mX2(y), mX3(z) {/* noop */}

template<class T>
inline StThreeVector<T>::~StThreeVector() {/* noop */}

template<class T>
inline void StThreeVector<T>::setX(T x) {mX1 = x;}

template<class T>
inline void StThreeVector<T>::setY(T y) {mX2 = y;}

template<class T>
inline void StThreeVector<T>::setZ(T z) {mX3 = z;}

template<class T>
void StThreeVector<T>::setPhi(T angle)
{
    double  r = magnitude();
    double th = theta();
    
    mX1 = r*sin(th)*cos(angle);
    mX2 = r*sin(th)*sin(angle);
}

template <class T>
void StThreeVector<T>::setTheta(T angle)
{
    double r  = magnitude();
    double ph = phi();

    mX1 = r*sin(angle)*cos(ph);
    mX2 = r*sin(angle)*sin(ph);
    mX3 = r*cos(angle);
}

template <class T>
void StThreeVector<T>::setMagnitude(T r)
{
    double th = theta();
    double ph = phi();
    
    mX1 = r*sin(th)*cos(ph);
    mX2 = r*sin(th)*sin(ph);
    mX3 = r*cos(th);
}

template <class T>
void StThreeVector<T>::setMag(T mag)
{
    setMagnitude(mag);
}

template<class T>
inline T StThreeVector<T>::x() const {return mX1;}

template<class T>
inline T StThreeVector<T>::y() const {return mX2;}

template<class T>
inline T StThreeVector<T>::z() const {return mX3;}

template<class T>
inline T StThreeVector<T>::theta() const
{
    return acos(mX3/this->mag());
}

template<class T>
inline T StThreeVector<T>::cosTheta() const
{
    return mX3/this->mag();
}

template<class T>
inline T StThreeVector<T>::phi() const
{
    return atan2(mX2,mX1);
}

template<class T>
inline T StThreeVector<T>::pseudoRapidity() const
{
    //
    // change code to more optimal:
    // double m = mag();
    // return 0.5*::log( (m+z())/(m-z()) );
    return -::log(tan(theta()/2.));
}

template<class T>
inline StThreeVector<T> StThreeVector<T>::unit() const
{
    return *this/this->mag();
}

template <class T>
T StThreeVector<T>::massHypothesis(T mass) const
{
    return ::sqrt((*this)*(*this) + mass*mass);
}

template <class T>
StThreeVector<T> StThreeVector<T>::orthogonal() const
{
    // Direct copy from CLHEP--it is probably better to
    // use your own dot/cross product code...
    double x = (mX1 < 0.0) ? -mX1 : mX1;
    double y = (mX2 < 0.0) ? -mX2 : mX2;
    double z = (mX3 < 0.0) ? -mX3 : mX3;
    
    if(x<y)
	return x < z ? StThreeVector<T>(0,mX3,-mX2) :  StThreeVector<T>(mX2,-mX1,0);
    else
	return  mX2 < mX3 ? StThreeVector<T>(-mX3,0,mX1) :  StThreeVector<T>(mX2,-mX1,0);
}

template <class T>
void StThreeVector<T>::rotateX(T angle)
{
    // may in the future make use of the StRotation class!
    double yPrime = cos(angle)*mX2 - sin(angle)*mX3;
    double zPrime = sin(angle)*mX2 + cos(angle)*mX3;

    mX2 = yPrime;
    mX3 = zPrime;
}

template <class T>
void StThreeVector<T>::rotateY(T angle)
{
    // may in the future make use of the StRotation class!
    double zPrime = cos(angle)*mX3 - sin(angle)*mX1;
    double xPrime = sin(angle)*mX3 + cos(angle)*mX1;

    mX1 = xPrime;
    mX3 = zPrime;
}

template <class T>
void StThreeVector<T>::rotateZ(T angle)
{
    // may in the future make use of the StRotation class!
    double xPrime = cos(angle)*mX1 - sin(angle)*mX2;
    double yPrime = sin(angle)*mX1 + cos(angle)*mX2;

    mX1 = xPrime;
    mX2 = yPrime;
}

template<class T>
inline T StThreeVector<T>::perp() const
{
    return ::sqrt(mX1*mX1+mX2*mX2);
}

template<class T>
inline T StThreeVector<T>::perp2() const
{
    return mX1*mX1+mX2*mX2;
}

template<class T>
inline T StThreeVector<T>::magnitude() const
{
    return mag();
}

template<class T>
inline T StThreeVector<T>::mag() const
{
    return ::sqrt(mX1*mX1+mX2*mX2+mX3*mX3);
}

template<class T>
inline T StThreeVector<T>::mag2() const
{
    return mX1*mX1+mX2*mX2+mX3*mX3;
}

template<class T>
inline T StThreeVector<T>::operator() (size_t i) const
{
    if (i == 0)
        return mX1;
    else if (i == 1)
        return mX2;
    else if (i == 2)
        return mX3;
    else {
      cerr << "StThreeVector<T>::operator(): bad index" << endl;
      return 0;
    }
}

template<class T>
inline T& StThreeVector<T>::operator() (size_t i)
{
    if (i == 0)
        return mX1;
    else if (i == 1)
        return mX2;
    else if (i == 2)
        return mX3;
    else {
      cerr << "StThreeVector<T>::operator(): bad index" << endl;
      return mX3;   // have to return something here ...
    }
}

template<class T>
inline T StThreeVector<T>::operator[] (size_t i) const
{
    if (i == 0)
        return mX1;
    else if (i == 1)
        return mX2;
    else if (i == 2)
        return mX3;
    else {
      cerr << "StThreeVector<T>::operator[]: bad index" << endl;
      return 0;
    }
}

template<class T>
inline T& StThreeVector<T>::operator[] (size_t i)
{
    if (i == 0)
        return mX1;
    else if (i == 1)
        return mX2;
    else if (i == 2)
        return mX3;
    else {
      cerr << "StThreeVector<T>::operator[]: bad index" << endl;
      return mX3;   // have to return something here ...
    }
}

template<class T>
inline StThreeVector<T>& StThreeVector<T>::operator*= (double c)
{
    mX1 *= c; mX2 *= c; mX3 *= c;
    return *this;
}

template<class T>
inline StThreeVector<T>& StThreeVector<T>::operator/= (double c)
{
    mX1 /= c; mX2 /= c; mX3 /= c;
    return *this;
}

template<class T>
inline StThreeVector<T>
StThreeVector<T>::pseudoProduct(double x,double y,double z) const
{
    return StThreeVector<T>(mX1*x,mX2*y,mX3*z);
}

template<class T>
StThreeVector<T> StThreeVector<T>::operator- ()
{
    return StThreeVector<T>(-mX1, -mX2, -mX3);
}

template<class T>
StThreeVector<T> StThreeVector<T>::operator+ ()
{
    return *this;
}

template<class T>
inline StThreeVector<T>::StThreeVector(const StThreeVector<float>& v)
    : mX1(v.x()), mX2(v.y()), mX3(v.z()) {/* noop */}

template<class T>
inline StThreeVector<T>::StThreeVector(const StThreeVector<double>& v)
    : mX1(v.x()), mX2(v.y()), mX3(v.z()) {/* noop */}

template<class T>
inline StThreeVector<T>::StThreeVector(const float *a)
{
    mX1 = a[0];
    mX2 = a[1];
    mX3 = a[2];
}

template<class T>
inline StThreeVector<T>::StThreeVector(const double *a)
{
    mX1 = a[0];
    mX2 = a[1];
    mX3 = a[2];
}

template<class T>
inline StThreeVector<T>&
StThreeVector<T>::operator=(const StThreeVector<float>& v)
{
    mX1 = v.x();  mX2 = v.y();  mX3 = v.z();
    return *this;
}

template<class T>
inline StThreeVector<T>&
StThreeVector<T>::operator=(const StThreeVector<double>& v)
{
    mX1 = v.x();  mX2 = v.y();  mX3 = v.z();
    return *this;
}

template<class T>
inline bool
StThreeVector<T>::operator== (const StThreeVector<float>& v) const
{
    return mX1 == v.x() && mX2 == v.y() && mX3 == v.z();
}

template<class T>
inline bool
StThreeVector<T>::operator== (const StThreeVector<double>& v) const
{
    return mX1 == v.x() && mX2 == v.y() && mX3 == v.z();
}

template<class T>
inline bool
StThreeVector<T>::operator!= (const StThreeVector<float>& v) const
{
    return !(*this == v);
}

template<class T>
inline bool
StThreeVector<T>::operator!= (const StThreeVector<double>& v) const
{
    return !(*this == v);
}

template<class T>
inline StThreeVector<T>&
StThreeVector<T>::operator+= (const StThreeVector<float>& v)
{
    mX1 += v.x(); mX2 += v.y(); mX3 += v.z();
    return *this;
}

template<class T>
inline StThreeVector<T>&
StThreeVector<T>::operator+= (const StThreeVector<double>& v)
{
    mX1 += v.x(); mX2 += v.y(); mX3 += v.z();
    return *this;
}

template<class T>
inline StThreeVector<T>&
StThreeVector<T>::operator-= (const StThreeVector<float>& v)
{
    mX1 -= v.x(); mX2 -= v.y(); mX3 -= v.z();
    return *this;
}

template<class T>
inline StThreeVector<T>&
StThreeVector<T>::operator-= (const StThreeVector<double>& v)
{
    mX1 -= v.x(); mX2 -= v.y(); mX3 -= v.z();
    return *this;
}

template<class T>
inline T StThreeVector<T>::dot(const StThreeVector<float>& v) const
{
    return mX1*v.x() + mX2*v.y() + mX3*v.z();
}

template<class T>
inline T StThreeVector<T>::dot(const StThreeVector<double>& v) const
{
    return mX1*v.x() + mX2*v.y() + mX3*v.z();
}

template<class T>
inline StThreeVector<T>
StThreeVector<T>::cross(const StThreeVector<float>& v) const
{
    return StThreeVector<T>(mX2*v.z() - mX3*v.y(),
			    mX3*v.x() - mX1*v.z(),
			    mX1*v.y() - mX2*v.x());
}

template<class T>
inline StThreeVector<T>
StThreeVector<T>::cross(const StThreeVector<double>& v) const
{
    return StThreeVector<T>(mX2*v.z() - mX3*v.y(),
			    mX3*v.x() - mX1*v.z(),
			    mX1*v.y() - mX2*v.x());
}

template<class T>
inline T StThreeVector<T>::angle(const StThreeVector<float>& v) const
{
    return acos(this->dot(v)/this->mag()/v.mag());
}

template<class T>
inline T StThreeVector<T>::angle(const StThreeVector<double>& v) const
{
    return acos(this->dot(v)/this->mag()/v.mag());
}

template<class T>
inline StThreeVector<T>
StThreeVector<T>::pseudoProduct(const StThreeVector<float>& v) const
{
    return this->pseudoProduct(v.x(),v.y(),v.z());
}

template<class T>
inline StThreeVector<T>
StThreeVector<T>::pseudoProduct(const StThreeVector<double>& v) const
{
    return this->pseudoProduct(v.x(),v.y(),v.z());
}
template<class T>
inline int
StThreeVector<T>::valid(double world) const
{
#if WIN32
#define finite _finite
#endif
  for (int i=0;i<3;i++) {
    if (!::finite((&mX1)[i])      ) return 0; 		
    if ( ::fabs  ((&mX1)[i])>world) return 0; 		
  }		
  return 1;		
}

//
//        Non-member functions
//
template<class T>
inline T abs(const StThreeVector<T>& v) {return v.mag();}

template<class T, class X>
inline StThreeVector<T>
cross_product(const StThreeVector<T>& v1, const StThreeVector<X>& v2)
{
    return v1.cross(v2);
}


//
//        Non-member operators
//
template<class T, class X>
inline StThreeVector<T>
operator+ (const StThreeVector<T>& v1, const StThreeVector<X>& v2)
{
    return StThreeVector<T>(v1) += v2;
}

template<class T, class X>
inline StThreeVector<T>
operator- (const StThreeVector<T>& v1, const StThreeVector<X>& v2)
{
    return StThreeVector<T>(v1) -= v2;
}

template<class T, class X>
inline T operator* (const StThreeVector<T>& v1, const StThreeVector<X>& v2)
{
    return StThreeVector<T>(v1).dot(v2);
}

template<class T>
inline StThreeVector<T> operator* (const StThreeVector<T>& v, double c)
{
    return StThreeVector<T>(v) *= c;
}

template<class T>
inline StThreeVector<T> operator* (double c, const StThreeVector<T>& v)
{
    return StThreeVector<T>(v) *= c;
}

template<class T, class X>
inline StThreeVector<T> operator/ (const StThreeVector<T>& v, X c)
{
    return StThreeVector<T>(v) /= c;
}

template<class T>
ostream&  operator<<(ostream& os, const StThreeVector<T>& v)
{
    return os << v.x() << '\t' << v.y() << '\t' << v.z();
}

template<class T>
istream&  operator>>(istream& is, StThreeVector<T>& v)
{
    T  x, y, z;
    is >> x >> y >> z;
    v.setX(x);
    v.setY(y);
    v.setZ(z);
    return is;
}
#endif
