//----------------------------------------------------------------------
//  Author: Thomas S. Ullrich (Feb 2005)
//
//  This file is part of hxBrowser.
//  
//  Description: Definitions of units taken from StarClassLibrary
//----------------------------------------------------------------------
#ifndef HEP_SYSTEM_OF_UNITS_H
#define HEP_SYSTEM_OF_UNITS_H

#if defined(WIN32)
#define M_PI 3.14159265358979323846
#endif

#include <cmath>

// 
// Length [L]
//
static const double millimeter  = 0.1;
static const double millimeter2 = millimeter*millimeter;
static const double millimeter3 = millimeter*millimeter*millimeter;

static const double centimeter  = 10*millimeter;
static const double centimeter2 = centimeter*centimeter;
static const double centimeter3 = centimeter*centimeter*centimeter;

static const double meter       = 100.*centimeter;
static const double meter2      = meter*meter;
static const double meter3      = meter*meter*meter;

static const double kilometer   = 1000.*meter;
static const double kilometer2  = kilometer*kilometer;
static const double kilometer3  = kilometer*kilometer*kilometer;

static const double micrometer  = 1.e-6*meter;
static const double nanometer   = 1.e-9*meter;
static const double femtometer  = 1.e-15*meter;
static const double fermi       = 1*femtometer;

static const double      barn   = 1.e-28*meter2;
static const double millibarn   = 1.e-3*barn;
static const double microbarn   = 1.e-6*barn;
static const double  nanobarn   = 1.e-9*barn;
static const double      inch   = 2.54*centimeter;

//
// Angle
//
static const double      radian = 1.;
static const double milliradian = 1.e-3*radian;
static const double      degree = (M_PI/180.0)*radian;
static const double   steradian = 1.;

//
// Time [T]
//
static const double      second = 1;
static const double millisecond = 1.e-3*second;
static const double microsecond = 1.e-3*millisecond;
static const double  nanosecond = 1.e-3*microsecond;

static const double     hertz   = 1./second;
static const double kilohertz   = 1.e+3*hertz;
static const double Megahertz   = 1.e+6*hertz;

// but these are also unambiguous and unlikely to be used as variable!
static const double  Hz         = 1*hertz;
static const double kHz         = 1*kilohertz;
static const double MHz         = 1*Megahertz;

//
// Electric charge [Q]
//
static const double eplus   = 1. ;		        // positron charge
static const double e_SI    = 1.60217733e-19;	// positron charge in coulomb
static const double coulomb = eplus/e_SI;

//
// Energy [E]
//
static const double Gigaelectronvolt = 1.;
static const double Megaelectronvolt = 1.e-3*Gigaelectronvolt;
static const double     electronvolt = 1.e-6*Megaelectronvolt;
static const double kiloelectronvolt = 1.e+3*electronvolt;
static const double Teraelectronvolt = 1.e+3*Gigaelectronvolt;

// but these are also unambiguous and unlikely to be used as variables
static const double MeV     = Megaelectronvolt;
static const double  eV     =     electronvolt;
static const double keV     = kiloelectronvolt;
static const double GeV     = Gigaelectronvolt;
static const double TeV     = Teraelectronvolt;

static const double joule   = electronvolt/e_SI;

//
// Mass [E][T^2][L^-2]
//
static const double  kilogram = joule*second*second/(meter*meter);
static const double      gram = 1.e-3*kilogram;
static const double milligram = 1.e-3*gram;

//
// Power [E][T^-1]
//
static const double watt    = joule/second;

//
// Force [E][L^-1]
//
static const double newton  = joule/meter;

//
// Pressure [E][L^-3]
//
static const double hep_pascal = newton/meter2;
static const double bar        = 100000*hep_pascal;
static const double atmosphere = 101325*hep_pascal;

//
// Electric current [Q][T^-1]
//
static const double ampere   = coulomb/second;

//
// Electric potential [E][Q^-1]
//
static const double Megavolt = MeV/eplus;
static const double kilovolt = 1.e-3*Megavolt;
static const double     volt = 1.e-6*Megavolt;
static const double millivolt = 1.e-3*volt;

//
// Electric resistance [E][T][Q^-2]
//
static const double ohm = volt/ampere;

//
// Electric capacitance [Q^2][E^-1]
//
static const double farad = coulomb/volt;
static const double millifarad = 1.e-3*farad;
static const double microfarad = 1.e-6*farad;
static const double  nanofarad = 1.e-9*farad;
static const double  picofarad = 1.e-12*farad;

//
// Magnetic Flux [T][E][Q^-1]
//
static const double weber = volt*second;

//
// Magnetic Field [T][E][Q^-1][L^-2]
    //
static const double tesla     = volt*second/meter2;

static const double gauss     = 1.e-4*tesla;
static const double kilogauss = 1.e-1*tesla;

//
// Inductance [T^2][E][Q^-2]
//
static const double henry = weber/ampere;

//
// Temperature
//
static const double kelvin = 1.;

//
// Amount of substance
    //
static const double mole = 1.;

//
// Activity [T^-1]
//
static const double becquerel = 1./second;
static const double curie = 3.7e+10 * becquerel;

//
// Absorbed dose [L^2][T^-2]
//
static const double gray = joule/kilogram ;

//
// Miscellaneous
//
static const double perCent     = 0.01 ;
static const double perThousand = 0.001;
static const double perMillion  = 0.000001;
#endif
