//----------------------------------------------------------------------
//  Author: Thomas S. Ullrich (Feb 2005)
//
//  This file is part of hxBrowser.
//  
//  Description: Implementation of hxAlgorithmWindowImp class
//----------------------------------------------------------------------
#include <qlineedit.h> 
#include <qstring.h>
#include <qmessagebox.h>
#include <qcheckbox.h>
#include <qtabwidget.h>
#include "hxAlgorithmWindowImp.h"
#include "hxCanvas.h"
#include "StPhysicalHelix.h"

hxAlgorithmWindowImp::hxAlgorithmWindowImp(QWidget* parent, const char* name, bool modal, WFlags fl) 
: hxAlgorithmWindow(parent, name, modal, fl)
{
    mHelix = 0;
    mCanvas = 0;
}

hxAlgorithmWindowImp::~hxAlgorithmWindowImp() {/* noop */}

void hxAlgorithmWindowImp::setHelix(StPhysicalHelix* helix)
{
    mHelix = helix;
    if (helix && isVisible()) {
        if (tabWidget->currentPageIndex() == 0)
            calculateDCAToPoint();
        else if (tabWidget->currentPageIndex() == 1)
            calculateIntersectionWithCylinder();
    }
}

void hxAlgorithmWindowImp::setCanvas(hxCanvas* canvas)
{
    mCanvas = canvas;
}

void hxAlgorithmWindowImp::calculateIntersectionWithCylinder()
{
    if (!mHelix) {
        QMessageBox::critical( this, "hxBrowser",
            "There is no helix defined yet. Please define the\n"
            "parameters of the helix in the main window first.");
        return;
    }

    QString text;
    bool isOK;

    double R = radiusField->text().toDouble(&isOK);
    if (!isOK || R < 0) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the radius of the cylinder is\n"
            "invalid. Please enter a valid positive numeric value.");
        return;
    }

    double xoff = offsetXField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the x offset of the cylinder is\n"
            "invalid. Please enter a valid numeric value.");
        return;
    }

    double yoff = offsetYField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the y offset of the cylinder is\n"
            "invalid. Please enter a valid numeric value.");
        return;
    }
 
    QString txt;
    txt.setNum(mHelix->period());
    periodField_2->setText(txt);

    pair<double, double> s = mHelix->pathLength(R, xoff, yoff);
    txt.setNum(s.first);
    pathLengthField1->setText(txt);
    txt.setNum(s.second);
    pathLengthField2->setText(txt);

    StThreeVector<> point1 = mHelix->at(s.first);
    StThreeVector<> point2 = mHelix->at(s.second);
    txt.setNum(point1.x());
    helixIntersectionPointX1->setText(txt);
    txt.setNum(point1.y());
    helixIntersectionPointY1->setText(txt);
    txt.setNum(point1.z());
    helixIntersectionPointZ1->setText(txt);
    txt.setNum(point2.x());
    helixIntersectionPointX2->setText(txt);
    txt.setNum(point2.y());
    helixIntersectionPointY2->setText(txt);
    txt.setNum(point2.z());
    helixIntersectionPointZ2->setText(txt);

    mCanvas->displayIntersectionWithCylinder(R, xoff, yoff); 
}

void hxAlgorithmWindowImp::calculateDCAToPoint()
{
    if (!mHelix) {
        QMessageBox::critical( this, "hxBrowser",
            "There is no helix defined yet. Please define the\n"
            "parameters of the helix in the main window first.");
        return;
    }

    QString text;
    bool isOK;

    double x = pointXField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the x position of the point\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double y = pointYField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the y position of the point\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double z = pointZField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the z position of the point\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    bool scanPeriods = ! limitToOnePeriodCheckBox->isChecked();
    StThreeVector<> point(x,y,z);

    QString txt;
    txt.setNum(mHelix->period());
    periodField->setText(txt);

    double s = mHelix->pathLength(point, scanPeriods);
    txt.setNum(s);
    pathLengthField->setText(txt);

    txt.setNum(abs(mHelix->at(s)-point));
    dcaField->setText(txt);

    StThreeVector<> hp = mHelix->at(s);
    txt.setNum(hp.x());
    helixXField->setText(txt);
    txt.setNum(hp.y());
    helixYField->setText(txt);
    txt.setNum(hp.z());
    helixZField->setText(txt);

    mCanvas->displayDCAToPoint(point, scanPeriods); 
}

void hxAlgorithmWindowImp::closeEvent(QCloseEvent *event) 
{
    mCanvas->removeDCAToPointDisplay();
    mCanvas->removeIntersectionWithCylinderDisplay(); 
    hxAlgorithmWindow::closeEvent(event);    
}

void hxAlgorithmWindowImp::tabChanged(QWidget*)
{
    int page = tabWidget->currentPageIndex();
    switch(page) {
        case 0:
            mCanvas->removeIntersectionWithCylinderDisplay(); 
            break;
        case 1:
            mCanvas->removeDCAToPointDisplay();
            break;
        default:                // oops
            mCanvas->removeDCAToPointDisplay();
            mCanvas->removeIntersectionWithCylinderDisplay(); 
            break;
    }
}


