//----------------------------------------------------------------------
//  Author: Thomas S. Ullrich (Feb 2005)
//
//  This file is part of hxBrowser.
//  
//  Description: Implementation of hxCanvas class
//----------------------------------------------------------------------
#include <iostream>
#include <qcursor.h>
#include "hxSettings.h"
#include "hxCanvas.h"
#include "SystemOfUnits.h"
#include "StPhysicalHelix.h"

using namespace std;

vector<size_t> hxCanvas::mDisplayLists(last,0);

hxCanvas::hxCanvas(QWidget *parent, const char *name )
: QGLWidget(parent, name) 
{
    setFormat(QGLFormat(DoubleBuffer|DepthBuffer));
    mTranslateX = 0;
    mTranslateY = 0;
    mTranslateZ = 0;
    mRotationX = 0;
    mRotationY = 0;
    mRotationZ = 0;
    mScale = 1;
    mShowTpcFrame = true;
    mShowBeamPipe = false;
    mShowAxis = false;
    mShowDCA = false;
    mShowCylinder = false;
    mHelix = 0;

    // dimensions in world corrdinates
    mWCSizeX = 6*meter;
    mWCSizeY = 6*meter;
    mWCSizeZ = 1000*meter;
}
   
hxCanvas::~hxCanvas() {/* noop */}    

void hxCanvas::setHelix(StPhysicalHelix* val)
{
    mHelix = val;

    // store helix in display list
    if (mDisplayLists[helix]) glDeleteLists(mDisplayLists[helix],1);
    mDisplayLists[helix] = glGenLists (1);
    glNewList(mDisplayLists[helix], GL_COMPILE);
    drawHelix();
    glEndList();

    updateGL();
}

void hxCanvas::initializeGL()
{
    // This virtual function is called once before the first call to 
    // paintGL() or resizeGL(), and then once whenever the widget has
    // been assigned a new QGLContext. Reimplement it in a subclass. 
    // This function should set up any required OpenGL context
    // rendering flags, defining display lists, etc. 
    // There is no need to call makeCurrent() because this has 
    // already been done when this function is called. 
    // Set up the rendering context, define display lists etc.:

    glClearColor( 0.0, 0.0, 0.0, 0.0 );
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glEnable(GL_COLOR_LOGIC_OP);
    glShadeModel(GL_FLAT);

    // store TPC in display list
    mDisplayLists[tpc] = glGenLists (1);
    glNewList(mDisplayLists[tpc], GL_COMPILE);
    drawTpcFrame();
    glEndList();

    // store beam pipe in display list
    mDisplayLists[beampipe] = glGenLists (1);
    glNewList(mDisplayLists[beampipe], GL_COMPILE);
    drawBeamPipe();
    glEndList();

    // store axis in display list
    mDisplayLists[axis] = glGenLists (1);
    glNewList(mDisplayLists[axis], GL_COMPILE);
    drawAxis();
    glEndList();
}

void hxCanvas::resizeGL( int w, int h )
{
    // This virtual function is called whenever the widget
    // has been resized. The new size is passed in width and height. 
    // Reimplement it in a subclass. 
    // There is no need to call makeCurrent() because this has already 
    // been done when this function is called. 
    // setup viewport, projection etc.:

    if (w < h)
        h = w;
    else if (w > h)
        w = h;
    glViewport(0, 0, (GLint) w, (GLint) h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-mWCSizeX/2, mWCSizeX/2, -mWCSizeX/2, mWCSizeY/2, -mWCSizeZ/2, mWCSizeZ/2);
	glMatrixMode(GL_MODELVIEW);	
    glLoadIdentity();
}

void hxCanvas::paintGL()
{
    // This virtual function is called whenever the widget needs 
    // to be painted. Reimplement it in a subclass. 
    // There is no need to call makeCurrent() because this has
    // already been done when this function is called. 
    // draw the scene:
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    draw();
}

void hxCanvas::draw() 
{
    //
    //   Here we draw everything at once using the
    //   predefined display lists.
    //
    glPushMatrix();      	// remember where we are        
    glScaled(mScale, mScale, mScale);
    glTranslated(mTranslateX, mTranslateY, mTranslateZ);
    glRotated(mRotationX, 1, 0, 0);
    glRotated(mRotationY, 0, 1, 0);
    glRotated(mRotationZ, 0, 0, 1);

    if (mShowTpcFrame) glCallList(mDisplayLists[tpc]);
    if (mShowBeamPipe) glCallList(mDisplayLists[beampipe]);
    if (mShowAxis)     glCallList(mDisplayLists[axis]);
    glCallList(mDisplayLists[helix]);
    if (mShowDCA)      glCallList(mDisplayLists[dca]);
    if (mShowCylinder) glCallList(mDisplayLists[cylinder]);

    glPopMatrix();      	// back to where we were
    glFlush();
 }

void hxCanvas::rotateX(int val)
{
    mRotationX = val;
    updateGL();
}
void hxCanvas::rotateY(int val)
{
    mRotationY = val;
    updateGL();
}
void hxCanvas::rotateZ(int val)
{
    mRotationZ = val;
    updateGL();
}

void hxCanvas::drawTpcFrame()
{
    double  tpcInnerRadius= 0.5*meter;
    double  tpcOuterRadius = 2.0*meter;
    double  tpcLength= 4.0*meter;

    const int beginAngle = 15;  // used to be zero
    const int nSectors = 12;
    
    double angleStep = 360./nSectors;
    double angle;
	
    glColor3d(0.3,0.3,0.3);	// dark grey
    glBegin(GL_LINES);
    for (angle = beginAngle; angle<360; angle+=angleStep) {     
		glVertex3d(tpcInnerRadius*cos(degree*angle), tpcInnerRadius*sin(degree*angle), -tpcLength/2.); 
		glVertex3d(tpcOuterRadius*cos(degree*angle), tpcOuterRadius*sin(degree*angle), -tpcLength/2.); 
		glVertex3d(tpcInnerRadius*cos(degree*angle), tpcInnerRadius*sin(degree*angle), tpcLength/2.); 
		glVertex3d(tpcOuterRadius*cos(degree*angle), tpcOuterRadius*sin(degree*angle), tpcLength/2.); 
		glVertex3d(tpcInnerRadius*cos(degree*angle), tpcInnerRadius*sin(degree*angle), -tpcLength/2.); 
		glVertex3d(tpcInnerRadius*cos(degree*angle), tpcInnerRadius*sin(degree*angle), tpcLength/2.); 
		glVertex3d(tpcOuterRadius*cos(degree*angle), tpcOuterRadius*sin(degree*angle), -tpcLength/2.); 
		glVertex3d(tpcOuterRadius*cos(degree*angle), tpcOuterRadius*sin(degree*angle), tpcLength/2.); 
    }

    glEnd();
    glBegin(GL_LINE_LOOP);
    for (angle = beginAngle; angle<360; angle+=angleStep) {     
		glVertex3d(tpcInnerRadius*cos(degree*angle), tpcInnerRadius*sin(degree*angle), -tpcLength/2.); 
    }
    glEnd();
    glBegin(GL_LINE_LOOP);
    for (angle = beginAngle; angle<360; angle+=angleStep) {     
		glVertex3d(tpcOuterRadius*cos(degree*angle), tpcOuterRadius*sin(degree*angle), -tpcLength/2.); 
    }
    glEnd();
    glBegin(GL_LINE_LOOP);
    for (angle = beginAngle; angle<360; angle+=angleStep) {     
		glVertex3d(tpcInnerRadius*cos(degree*angle), tpcInnerRadius*sin(degree*angle), tpcLength/2.); 
    }
    glEnd();
    glBegin(GL_LINE_LOOP);
    for (angle = beginAngle; angle<360; angle+=angleStep) {     
		glVertex3d(tpcOuterRadius*cos(degree*angle), tpcOuterRadius*sin(degree*angle), tpcLength/2.); 
    }
    glEnd();
}

void hxCanvas::drawBeamPipe()
{
    const double radius = 5.0*centimeter;
    const double length = 4.5*meter;
	const double gray = 0.3;

    static GLUquadricObj *qobj = 0;

    glColor3d(gray, gray, gray);
	if (qobj) gluDeleteQuadric(qobj);
	qobj = gluNewQuadric();
    gluQuadricDrawStyle(qobj, GLU_LINE);
	gluCylinder(qobj, radius, radius, length/2, 36, 20);
	glTranslatef(0, 0, -length/2);
	gluCylinder(qobj, radius, radius, length/2, 36, 20);
	glTranslatef(0, 0, length/2);
}

void hxCanvas::drawAxis()
{
    double xlen = meter;
    double ylen = meter;
    double zlen = meter;
    double txtsize = 8*centimeter;
	    	
	// Coordinate system
	glColor3d(1.,0.,0.);	// red
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(xlen, 0.0, 0.0);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, ylen, 0.0);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, 0.0, zlen);
	glEnd();
	
	// axis labels
	glColor3d(1.,1.,0.); 	// yellow 
	drawText(1.05*xlen, 0, 0, "x", txtsize); 
	drawText(0, 1.05*ylen, 0, "y", txtsize);    
	drawText(0, 0, 1.05*zlen, "z", txtsize);   
}

void hxCanvas::drawText(double x, double y, double z, char *txt, double size)
{
    // This was originally done with:
    // glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
    // However, this was the only GLUT function used and
    // glut is not as readily available as OpenGL.
    // Therefore we draw the characters by hand.
    // The function accepts only x,y,z as characters.
    // The character height is roughly "size" (5th arg)
    // in centimeter.

    glPushMatrix();
    glTranslated(x, y, z);
    glScaled(size, size, size);
    glRotated(-mRotationX, 1.0, 0.0, 0.0);
    glRotated(-mRotationY, 0.0, 1.0, 0.0);
    glRotated(-mRotationZ, 0.0, 0.0, 1.0);
    //    for (char *p = txt; *p; p++)
    //		  glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
    if (*txt == 'y') {
        glBegin(GL_LINES);
        glVertex3d(0,0,0); 
        glVertex3d(-0.5,1,0); 
        glVertex3d(-0.5,-1,0); 
        glVertex3d(0.5,1,0); 
        glEnd();
    }
    else if (*txt == 'x') {
        glBegin(GL_LINES);
        glVertex3d(-0.5,-0.625,0); 
        glVertex3d(0.5,0.625,0); 
        glVertex3d(0.5,-0.625,0); 
        glVertex3d(-0.5,0.625,0); 
        glEnd();
    }
    else if (*txt == 'z') {
        glBegin(GL_LINE_STRIP);
        glVertex3d(-0.5,0.625,0); 
        glVertex3d(0.5,0.625,0); 
        glVertex3d(-0.5,-0.625,0); 
        glVertex3d(0.5,-0.625,0); 
        glEnd();
    }
    glPopMatrix();
}

void hxCanvas::drawHelix()
{
    if (!mHelix) return;

    hxSettings* settings = hxSettings::instance();

    glBegin(GL_LINE_STRIP);
    glColor3d(0, 0, 1);                  // blue   
    
    StThreeVector<> p;
	for (double s = 0; s<settings->mMaxHelixLength.value(); s+=settings->mHelixDrawingSegmentSize.value()) {
		p = mHelix->at(s);
		if (fabs(p.z()) > settings->mMaxZForHelixDrawing.value()) break;
		if (p.perp() > settings->mMaxRForHelixDrawing.value()) break;
		glVertex3d(p.x(), p.y(), p.z());
	}

	glEnd();
}

void hxCanvas::setShowTpcFrame(bool val)
{
    mShowTpcFrame = val;   
    updateGL();
}

void hxCanvas::setShowBeamPipe(bool val) 
{
    mShowBeamPipe = val;   
    updateGL();
}

void hxCanvas::setShowAxis(bool val)
{
    mShowAxis = val;   
    updateGL();
}

void hxCanvas::resetZoomAndShift()
{
    mTranslateX = 0;
    mTranslateY = 0;
    mTranslateZ = 0;
    mScale = 1;
    updateGL();
}

void hxCanvas::keyPressEvent(QKeyEvent* event)
{    
    bool altPressed = event->state() == AltButton;
    bool shiftPressed = event->state() == ShiftButton;
    bool ctrlPressed = event->state() == ControlButton;

    const double tStep =  centimeter;
    const double fast_tStep = 4*centimeter;
    const double scaleFactor = 1.02;

    switch (event->key()) {
        case Key_Z:
            if (altPressed || shiftPressed)
                mScale /= scaleFactor;
            else
                mScale *= scaleFactor;
            break;
        case Key_R:
            if (ctrlPressed) 
                resetZoomAndShift();
            else
                mTranslateX += (shiftPressed ? fast_tStep : tStep);
            break;
        case Key_L:
            mTranslateX -= (shiftPressed ? fast_tStep : tStep);
            break;
        case Key_U:
            mTranslateY += (shiftPressed ? fast_tStep : tStep);
            break;
        case Key_D:
            mTranslateY -= (shiftPressed ? fast_tStep : tStep);
            break;
        case Key_F:
            mTranslateZ += (shiftPressed ? fast_tStep : tStep);
            break;
        case Key_B:
            mTranslateZ -= (shiftPressed ? fast_tStep : tStep);
            break;
        default:
            QGLWidget::keyPressEvent(event);
            break;
    }
    updateGL();
}

void hxCanvas::mousePressEvent(QMouseEvent* event) 
{
    if (event->button() == RightButton) {
        mLastMousePosition = event->pos();
        setCursor(SizeAllCursor);
    }
    else
        QGLWidget::mousePressEvent(event);
}

void hxCanvas::mouseReleaseEvent(QMouseEvent* event) 
{
    if (event->button() == RightButton)
        unsetCursor();
    else
        QGLWidget::mouseReleaseEvent(event);
}

void hxCanvas::mouseMoveEvent(QMouseEvent* event) 
{
    if (event->state() == RightButton) {
        double dx = static_cast<double>((event->x() - mLastMousePosition.x()))/width(); 
        double dy = static_cast<double>((event->y() - mLastMousePosition.y()))/height();
        mTranslateX += mWCSizeX*dx/mScale;
        mTranslateY -= mWCSizeX*dy/mScale;
        mLastMousePosition = event->pos();
        updateGL();
    }
    else
        QGLWidget::mouseMoveEvent(event);
}

void hxCanvas::wheelEvent(QWheelEvent* event) 
{
    const double scaleStep = 1.05;

    if (event->delta() > 0) {
        mScale *= scaleStep;
        updateGL();
    }
    else if (event->delta() < 0) {
        mScale /= scaleStep;
        updateGL();
    }
    else
        QGLWidget::wheelEvent(event);
}

void hxCanvas::displayDCAToPoint(StThreeVector<>& point, bool scanPeriods) 
{
    // store DCA lines in display list
    if (mDisplayLists[dca]) glDeleteLists(mDisplayLists[dca],1);
    mDisplayLists[dca] = glGenLists(1);
    glNewList(mDisplayLists[dca], GL_COMPILE);

    double xlen = centimeter;
    double ylen = centimeter;
    double zlen = centimeter;

	glColor3d(1.,1.,0.); 	// yellow 
	glBegin(GL_LINES);
	glVertex3d(point.x()-xlen/2, point.y(), point.z());
	glVertex3d(point.x()+xlen/2, point.y(), point.z());
	glVertex3d(point.x(), point.y()-ylen/2, point.z());
	glVertex3d(point.x(), point.y()+ylen/2, point.z());
	glVertex3d(point.x(), point.y(), point.z()-zlen/2);
	glVertex3d(point.x(), point.y(), point.z()+zlen/2);
	glEnd();

    glColor3d(0.,1.,0.); 	// green 
	glBegin(GL_LINES);
    double s = mHelix->pathLength(point, scanPeriods);
    StThreeVector<> v = mHelix->at(s);
	glVertex3d(point.x(), point.y(), point.z());
	glVertex3d(v.x(), v.y(), v.z());
	glEnd();
    glEndList();

    mShowDCA = true;
    updateGL();
}

void hxCanvas::displayIntersectionWithCylinder(double R, double x, double y)
{
    if (mDisplayLists[cylinder]) glDeleteLists(mDisplayLists[cylinder],1);
    mDisplayLists[cylinder] = glGenLists(1);
    glNewList(mDisplayLists[cylinder], GL_COMPILE);

    pair<double, double> s  = mHelix->pathLength(R, x, y);
    StThreeVector<> point1 = mHelix->at(s.first);
    StThreeVector<> point2 = mHelix->at(s.second);

    hxSettings *settings = hxSettings::instance();
    double xlen = settings->m3DPointSegmentLength.value();
    double ylen = xlen;
    double zlen = xlen;

    glColor3d(1.,1.,0.); 	// yellow 
    glBegin(GL_LINES);
    glVertex3d(point1.x()-xlen/2, point1.y(), point1.z());
    glVertex3d(point1.x()+xlen/2, point1.y(), point1.z());
    glVertex3d(point1.x(), point1.y()-ylen/2, point1.z());
    glVertex3d(point1.x(), point1.y()+ylen/2, point1.z());
    glVertex3d(point1.x(), point1.y(), point1.z()-zlen/2);
    glVertex3d(point1.x(), point1.y(), point1.z()+zlen/2);
    glVertex3d(point2.x()-xlen/2, point2.y(), point2.z());
    glVertex3d(point2.x()+xlen/2, point2.y(), point2.z());
    glVertex3d(point2.x(), point2.y()-ylen/2, point2.z());
    glVertex3d(point2.x(), point2.y()+ylen/2, point2.z());
    glVertex3d(point2.x(), point2.y(), point2.z()-zlen/2);
    glVertex3d(point2.x(), point2.y(), point2.z()+zlen/2);
    glEnd();

    const double length = settings->mCylinderLength.value();
    const int slices = settings->mNumberOfCylinderSlices.value();
    const int stacks = settings->mNumberOfCylinderStacks.value();

    static GLUquadricObj *qobj = 0;

    glTranslated(x, y, 0);
    glColor3d(0.0, 1.0, 0.0);   // green
    if (qobj) gluDeleteQuadric(qobj);
    qobj = gluNewQuadric();
    gluQuadricDrawStyle(qobj, GLU_LINE); 
    gluCylinder(qobj, R, R, length/2, slices, stacks);
    glTranslated(0, 0, -length/2);
    gluCylinder(qobj, R, R, length/2, slices, stacks);
    glTranslated(0, 0, length/2);
    glTranslated(-x, -y, 0);

    glEndList();

    mShowCylinder = true;
    updateGL();
}
    
void hxCanvas::removeDCAToPointDisplay()
{
    mShowDCA = false;
    updateGL();
}

void hxCanvas::removeIntersectionWithCylinderDisplay()
{
    mShowCylinder = false;
    updateGL();
}


