//----------------------------------------------------------------------
//  Author: Thomas S. Ullrich (Feb 2005)
//
//  This file is part of hxBrowser.
//  
//  Description: Implementation of hxMainWindowImp class
//----------------------------------------------------------------------
#include <qlineedit.h> 
#include <qstring.h>
#include <qmessagebox.h>
#include <qradiobutton.h> 
#include <qslider.h> 
#include <qaction.h>
#include <qlabel.h>
#include "hxMainWindowImp.h"
#include "hxAlgorithmWindowImp.h"
#include "hxSettingsDialog.h"
#include "hxHelpBrowser.h"
#include "hxCanvas.h"
#include "hxVersion.h"
#include "SystemOfUnits.h"

hxMainWindowImp::hxMainWindowImp( QWidget* parent, const char* name, WFlags fl) :
hxMainWindow(parent, name, fl) 
{
    mHelix = 0;
    mAlgorithmWindow = new hxAlgorithmWindowImp(this);
    mAlgorithmWindow->setCanvas(hxCanvasWidget);   
    this->viewAlgorithmsAction->setEnabled(false);
    mSettingsDialog = new hxSettingsDialog(this);
};

hxMainWindowImp::~hxMainWindowImp() 
{
    if (mHelix) delete mHelix;
    delete mAlgorithmWindow;
    delete mSettingsDialog;
};

void hxMainWindowImp::showAlgorithmWindow()
{
    mAlgorithmWindow->show();
}

void hxMainWindowImp::applyPhysicalHelix()
{
    //
    // Called when a new physical helix is defined
    // or parameters have changed.
    //

    QString text;
    bool isOK;

    double B = PTeslaField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the magnetic field (B) is invalid.\n"
            "Please enter a valid numeric value.");
        return;
    }
    double x = POriginXField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the x position of the helix origin\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double y = POriginYField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the y position of the helix origin\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double z = POriginZField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the z position of the helix origin\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double pt = PPtField->text().toDouble(&isOK);
    if (!isOK || pt < 0) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the transverse momentum (pt) of the\n"
            "helix is invalid. Please enter a positive numeric value.");
        return;
    }
    double eta = PEtaField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the pseudo-rapidity (eta) of the\n"
            "helix is invalid. Please enter a valid numeric value.");
        return;
    }
    double phi = PPhiField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the azimuthal angle (phi) of the\n"
            "helix is invalid. Please enter a valid numeric value.");
        return;
    }

    int charge = PPositiveChargeButton->isChecked() ? 1 : -1;
    
    //
    // Create new physical helix from what we got
    // 
    if (hxSettings::instance()->mAnglesInDegree.value()) phi *= degree;

    StThreeVector<> origin(x,y,z);
	StThreeVector<> p(pt*cos(phi), pt*sin(phi), pt*sinh(eta));
    if (mHelix) delete mHelix;
	mHelix = new StPhysicalHelix(p, origin, B*tesla, charge);
    this->viewAlgorithmsAction->setEnabled(true);
    displayMathematicalHelixParameters();  // sync with mathematical helix

    hxCanvasWidget->setHelix(mHelix);
    mAlgorithmWindow->setHelix(mHelix);
    hxCanvasWidget->update();
}

void hxMainWindowImp::applyMathematicalHelix()
{
    //
    // Called when a new mathematical helix is defined
    // or parameters have changed.
    //

    QString text;
    bool isOK;

    double x = MOriginXField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the x position of the helix origin\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double y = MOriginYField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the y position of the helix origin\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double z = MOriginZField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the z position of the helix origin\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double curvature = MCurvatureField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the curvature of the helix\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
     if (curvature < 0) {
        QMessageBox::warning( this, "hxBrowser",
            "You entered a negative curvature. In the present parametrization\n"
            "this is not foreseen. The helicity is stored in a separate parameter.\n"
            "However, StHelix will transform all given parameters to end up with the\n"
            "equivalient parameter set with positive curvature. You've been warned!");
    }
    double dipAngle = MDipAngleField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the dip angle of the helix\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }
    double phase = MPhaseField->text().toDouble(&isOK);
    if (!isOK) {
        QMessageBox::critical( this, "hxBrowser",
            "The value given for the phase of the helix\n"
            "is invalid. Please enter a valid numeric value.");
        return;
    }

    int h = MPositiveHelicityButton->isChecked() ? 1 : -1;

    //
    // Create mathematical physical helix from what we got
    // 
    if (hxSettings::instance()->mAnglesInDegree.value()) {
        dipAngle *= degree;
        phase *= degree;
    }
    StThreeVector<> origin(x,y,z);
    if (mHelix) delete mHelix;
	mHelix = new StPhysicalHelix(curvature, dipAngle, phase, origin, h);
    this->viewAlgorithmsAction->setEnabled(true);

    displayPhysicalHelixParameters();  // sync with physical helix

    hxCanvasWidget->setHelix(mHelix);
    mAlgorithmWindow->setHelix(mHelix);
    hxCanvasWidget->update();

}

void hxMainWindowImp::displayPhysicalHelixParameters()
{    
    bool isOK;
    double B = PTeslaField->text().toDouble(&isOK);

    if (mHelix) {
        QString txt;
        txt.setNum(mHelix->origin().x());
        POriginXField->setText(txt);
        txt.setNum(mHelix->origin().y());
        POriginYField->setText(txt);
        txt.setNum(mHelix->origin().z());
        POriginZField->setText(txt);
        StThreeVector<> p = mHelix->momentum(B*tesla); 
        txt.setNum(p.perp());
        PPtField->setText(txt);
        txt.setNum(p.pseudoRapidity());
        PEtaField->setText(txt);
        double phi = p.phi();
        if (hxSettings::instance()->mAnglesInDegree.value()) phi /= degree;
        txt.setNum(phi);
        PPhiField->setText(txt);
        int charge = mHelix->charge(B*tesla);
        PNegativeChargeButton->setChecked(charge < 0 ? true : false);
        PPositiveChargeButton->setChecked(charge > 0 ? true : false);
    }
}

void hxMainWindowImp::displayMathematicalHelixParameters()
{
    if (mHelix) {
        QString txt;
        txt.setNum(mHelix->origin().x());
        MOriginXField->setText(txt);
        txt.setNum(mHelix->origin().y());
        MOriginYField->setText(txt);
        txt.setNum(mHelix->origin().z());
        MOriginZField->setText(txt);
        txt.setNum(mHelix->curvature());
        MCurvatureField->setText(txt);
        double dipAngle = mHelix->dipAngle();
        if (hxSettings::instance()->mAnglesInDegree.value()) dipAngle /= degree;
        txt.setNum(dipAngle);
        MDipAngleField->setText(txt);
        double phase = mHelix->phase();
        if (hxSettings::instance()->mAnglesInDegree.value()) phase /= degree;
        txt.setNum(phase);
        MPhaseField->setText(txt);
        MNegativeHelicityButton->setChecked(mHelix->h() > 0 ? false : true);
        MPositiveHelicityButton->setChecked(mHelix->h() < 0 ? false : true);
    }
}

void hxMainWindowImp::keyPressEvent(QKeyEvent* event)
{
    int pos;
    if (hxCanvasWidget->hasMouse()) {
        bool shiftPressed = event->state() == ShiftButton;

        switch (event->key()) {
        case Key_Left:
            pos = rotateYSlider->value()-1;
            if (pos < rotateYSlider->minValue()) 
                pos = rotateYSlider->minValue();
            rotateYSlider->setValue(pos);   
            break;
        case Key_Right:
            pos = rotateYSlider->value()+1;
            if (pos > rotateYSlider->maxValue()) 
                pos = rotateYSlider->maxValue();
            rotateYSlider->setValue(pos);   
            break;
        case Key_Down:
            if (shiftPressed) {
                pos = rotateZSlider->value()+1;
                if (pos > rotateZSlider->maxValue()) 
                    pos = rotateZSlider->maxValue();
                rotateZSlider->setValue(pos);
            }
            else {
                pos = rotateXSlider->value()+1;
                if (pos > rotateXSlider->maxValue()) 
                    pos = rotateXSlider->maxValue();
                rotateXSlider->setValue(pos);
            }
            break;
        case Key_Up:
            if (shiftPressed) {
                pos = rotateZSlider->value()-1;
                if (pos < rotateZSlider->minValue()) 
                    pos = rotateZSlider->minValue();
                rotateZSlider->setValue(pos);
            }
            else {
                pos = rotateXSlider->value()-1;
                if (pos < rotateXSlider->minValue()) 
                    pos = rotateXSlider->minValue();
                rotateXSlider->setValue(pos);   
            }
            break;
        default:
            hxMainWindow::keyPressEvent(event);
        }
    }
    else
        hxMainWindow::keyPressEvent(event);
}

void hxMainWindowImp::helpAbout()
{
    QMessageBox::about( this, "hxBrowser",
                        "hxBrowser\t Version " VERSION "\n" 
                        "A tool to visualize helices and test their implementation including\n"
                        "various algorithms - or simply to learn what a helix is all about.\n\n"
                        "This program is free software; you can redistribute it and/or modify\n"
                        "it under the terms of the GNU General Public License as published by\n"
                        "the Free Software Foundation.\n\n"
						"The program is distributed in the hope that it will be useful,\n"
						"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
						"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
						"See the GNU General Public License for more details.\n\n"
                        "Copyright 2005 Thomas S. Ullrich (thomas.ullrich@bnl.gov)\n");
    return;
}

void hxMainWindowImp::helpContents()
{
    hxHelpBrowser::showPage("index.html");
}

void hxMainWindowImp::showSettings()
{
    mSettingsDialog->show();
}

void hxMainWindowImp::syncWithSettings() 
{
    hxSettings *settings = hxSettings::instance();
    if (settings->mAnglesInDegree.value()) {
        angleLabel1->setText("deg");
        angleLabel2->setText("deg");
        angleLabel3->setText("deg");
    }
    else {
        angleLabel1->setText("rad");
        angleLabel2->setText("rad");
        angleLabel3->setText("rad");
    }
    displayMathematicalHelixParameters();
    displayPhysicalHelixParameters();
}
